#!/usr/bin/env python3
from __future__ import (unicode_literals, absolute_import,
                        print_function, division)

import collections
import inspect
import rlcompleter
import re, sys
    
try:
    # Python >= 3.3
    from importlib.machinery import all_suffixes
    _suffixes = all_suffixes()
except ImportError:
    from imp import get_suffixes
    _suffixes = [ s[0] for s in get_suffixes() ]

# Regular expressions for module aliases e.g. “np” -> “numpy”:
from pythonpy.pyeval import alias_res
    
# Regular expression for the python import statement
import_re = re.compile(r'(?P<name>[a-zA-Z_][a-zA-Z0-9_]*?)'
                       r'(?P<package>[/\\]__init__)?'
                       r'(?P<suffix>%s)$' %
                       r'|'.join(re.escape(s) for s in _suffixes))

def current_raw(input):
    if len(input[-1]) > 0 and input[-1][0] in '"\'':
        return input[-1][1:]
    return input[-1]

def current_list(input):
    return current_list.rgx.split(current_raw(input))

current_list.rgx = re.compile(r'[^a-zA-Z0-9_\.]')

def current_prefix(input):
    return current_list(input)[-1]

def prior(input):
    return input[:-1]

def lazy_imports(*args):
    query = ' '.join([x for x in args if x])
    matches = lazy_imports.rgx.findall(query)
    
    for raw_module_name in matches:
        
        module_name = raw_module_name
        for rgx, alias in alias_res.items():
            if rgx.match(module_name):
                module_name = rgx.sub(alias.modname, module_name)
        
        try:
            module = __import__(module_name)
        except ImportError:
            pass
        else:
            globals()[raw_module_name] = module

lazy_imports.rgx = re.compile("([a-zA-Z_][a-zA-Z0-9_]*)\.?")

def complete_all(prefix, completion_args):
    lazy_imports(prefix, completion_args['c_arg'])
    if completion_args:
        if completion_args['x_arg']:
            x = str()
        if completion_args['l_arg']:
            l = list()
        if completion_args['c_arg']:
            exec(completion_args['c_arg'].strip('"\'').replace("`", "'"))
    context = locals()
    context.update(globals())
    completer = rlcompleter.Completer(context)
    idx = 0
    options_set = set()
    
    while completer.complete(prefix, idx):
        options_set.add(completer.complete(prefix, idx))
        idx += 1
    
    module_completion, module_list = get_completerlib()
    try:
        options = module_completion("import " + prefix) or []
    except:
        # module_completion may throw exception (e.g. on 'import sqlalchemy_utils.')
        options = []
    if options:
        options = [x.rstrip(' ') for x in options if x.startswith(prefix)]
    
    return options + list(options_set)

def parse_string(input):
    if current_raw(input).startswith('--'):
        return ['--si', '--so', '--ji', '--jo', '--i']
    elif current_raw(input).startswith('-'):
        return ['-h', '-x', '-fx', '-l', '-c', '-C']
    elif len(prior(input)) > 0 and prior(input)[-1] == '-c':
        if 'import'.startswith(current_raw(input)):
            options = ["'import"]
        elif current_raw(input).startswith('import ') or current_raw(input).startswith('from '):
            module_completion, module_list = get_completerlib()
            options = module_completion(current_raw(input)) or []
            if options:
                options = [x.rstrip(' ') for x in options if x.startswith(current_prefix(input))]
        else:
            options = complete_all(current_prefix(input), collections.defaultdict(lambda: None))
            if current_prefix(input).endswith('.'):
                options = [x for x in options if '._' not in x]
        return options
    elif current_raw(input) == '':
        options = ['sys', 'json', 're', 'csv', 'datetime', 'hashlib', 'itertools', 'math', 'os', 'random', 'shutil'] 
        if '-x' in input[:-1] or '-fx' in input[:-1]:
            options += 'x'
        if '-l' in input[:-1]:
            options += 'l'
        return options
    else:
        completion_args = collections.defaultdict(lambda: None)
        if '-x' in prior(input) or '-fx' in prior(input):
            completion_args['x_arg'] = True
        if '-l' in prior(input):
            completion_args['l_arg'] = True
        if '-c' in prior(input):
            c_index = prior(input).index('-c')
            if (c_index + 1) < len(prior(input)):
                completion_args['c_arg'] = prior(input)[c_index + 1]
        options = complete_all(current_prefix(input), completion_args)
        if current_prefix(input).endswith('.'):
            options = [x for x in options if '._' not in x]
        return options

# RE for the ipython %run command (python + ipython scripts)
magic_run_re = re.compile(r'.*(\.ipy|\.ipynb|\.py[w]?)$')
    
def is_importable(module, attr, only_modules):
    if only_modules:
        return inspect.ismodule(getattr(module, attr))
    else:
        return not(attr[:2] == '__' and attr[-2:] == '__')

def get_completerlib():
    """Implementations for various useful completers.
    
    These are all loaded by default by IPython.
    """
    #-----------------------------------------------------------------------------
    #  Copyright (C) 2010-2011 The IPython Development Team.
    #
    #  Distributed under the terms of the BSD License.
    #
    #  The full license is in the file COPYING.txt, distributed with this software.
    #-----------------------------------------------------------------------------
    
    #-----------------------------------------------------------------------------
    # Imports
    #-----------------------------------------------------------------------------
    import os
    
    # Third-party imports
    from time import time
    from zipimport import zipimporter
    
    TIMEOUT_STORAGE = 2
    TIMEOUT_GIVEUP = 20
    
    def module_list(path):
        """
        Return the list containing the names of the modules available in the given
        folder.
        """
        # sys.path has the cwd as an empty string, but isdir/listdir need it as '.'
        if path == '':
            path = '.'
        
        # A few local constants to be used in loops below
        pjoin = os.path.join
        
        if os.path.isdir(path):
            # Build a list of all files in the directory and all files
            # in its subdirectories. For performance reasons, do not
            # recurse more than one level into subdirectories.
            files = []
            for root, dirs, nondirs in os.walk(path):
                subdir = root[len(path)+1:]
                if subdir:
                    files.extend(pjoin(subdir, f) for f in nondirs)
                    dirs[:] = [] # Do not recurse into additional subdirectories.
                else:
                    files.extend(nondirs)
        
        else:
            try:
                files = list(zipimporter(path)._files.keys())
            except:
                files = []
        
        # Build a list of modules which match the import_re regex.
        modules = []
        for f in files:
            m = import_re.match(f)
            if m:
                modules.append(m.group('name'))
        return list(set(modules))
    
    def get_root_modules():
        """
        Returns a list containing the names of all the modules available in the
        folders of the pythonpath.
        
        ip.db['rootmodules_cache'] maps sys.path entries to list of modules.
        """
        #ip = get_ipython()
        #rootmodules_cache = ip.db.get('rootmodules_cache', {})
        rootmodules_cache = {}
        rootmodules = list(sys.builtin_module_names)
        start_time = time()
        store = False
        for path in sys.path:
            
            try:
                modules = rootmodules_cache[path]
            except KeyError:
                modules = module_list(path)
                
                try:
                    modules.remove('__init__')
                except ValueError:
                    pass
                
                if path not in ('', '.'): # cwd modules should not be cached
                    rootmodules_cache[path] = modules
                
                if time() - start_time > TIMEOUT_STORAGE and not store:
                    #store = True
                    #print("\nCaching the list of root modules, please wait!")
                    #print("(This will only be done once - type '%rehashx' to "
                          #"reset cache!)\n")
                    sys.stdout.flush()
                if time() - start_time > TIMEOUT_GIVEUP:
                    print("This is taking too long, we give up.\n")
                    return []
            
            rootmodules.extend(modules)
        #if store:
            #ip.db['rootmodules_cache'] = rootmodules_cache
        rootmodules = list(set(rootmodules))
        return rootmodules
    
    def try_import(mod, only_modules=False):
        try:
            m = __import__(mod)
        except:
            return []
        
        mods = mod.split('.')
        
        for module in mods[1:]:
            m = getattr(m, module)
        
        m_is_init = hasattr(m, '__file__') and '__init__' in m.__file__
        
        completions = []
        
        if (not hasattr(m, '__file__')) or (not only_modules) or m_is_init:
            completions.extend( [attr for attr in dir(m) if
                                 is_importable(m, attr, only_modules)])
        
        completions.extend(getattr(m, '__all__', []))
        
        if m_is_init:
            completions.extend(module_list(os.path.dirname(m.__file__)))
        
        completions = set(completions)
        
        if '__init__' in completions:
            completions.remove('__init__')
        
        return list(completions)
    
    def module_completion(line):
        """
        Returns a list containing the completion possibilities for an import line.
        
        The line looks like this :
        'import xml.d'
        'from xml.dom import'
        """
        
        words = line.split(' ')
        nwords = len(words)
        
        # from whatever <tab> -> 'import '
        if nwords == 3 and words[0] == 'from':
            return ['import ']
        
        # 'from xy<tab>' or 'import xy<tab>'
        if nwords < 3 and (words[0] in ['import', 'from']) :
            if nwords == 1:
                return get_root_modules()
            mod = words[1].split('.')
            if len(mod) < 2:
                return get_root_modules()
            completion_list = try_import('.'.join(mod[:-1]), True)
            return ['.'.join(mod[:-1] + [el]) for el in completion_list]
        
        # 'from xyz import abc<tab>'
        if nwords >= 3 and words[0] == 'from':
            mod = words[1]
            return try_import(mod)
    
    return module_completion, module_list

def remove_trailing_paren(string):
    if string.endswith('('):
        return string[:-1]
    return string

def main():
    input = sys.argv[1:]
    if len(input) == 0:
        return
    elif '<' in input or '>' in input:
        print('_longopt')
        return
    else:
        options = list(set(map(remove_trailing_paren, parse_string(input))))
        
        if len(options) == 0:
            return
        
        if len(current_list(input)) > 1 and max(map(len, options)) + 1 >= len(current_raw(input)):
            options.append(current_prefix(input))
        
        if len(options) <= 1:
            options = options + [x + "'" for x in options]
        print(' '.join(options))

if __name__ == '__main__':
    main()




import unittest
from pythonpy.pyeval import pyeval

class TestPyEval(unittest.TestCase):
    
    def test_empty(self):
        self.assertEqual(pyeval([]), '')
    
    def test_numbers(self):
        self.assertEqual(pyeval(['3 * 4.5']), '13.5\n')
    
    def test_range(self):
        self.assertEqual(pyeval(['range(3)']), '0\n1\n2\n')
    
    def test_statements(self):
        self.assertEqual(pyeval(['-c', 'a=5', '-C', 'print(a)']), '5\n')
    
    def test_imports(self):
        module_commands = [
            "math.ceil(2.5)",
            "calendar.weekday(1955, 11, 5)",
            "csv.list_dialects()",
            "datetime.timedelta(hours=-5)",
            "hashlib.sha224(b\"Nobody inspects the spammish repetition\").hexdigest()",
            "glob.glob('*')",
            "itertools.product(['a','b'], [1,2])",
            "json.dumps([1,2,3,{'4': 5, '6': 7}], separators=(',',':'))",
            "os.name",
            "random.randint(0, 1000)",
            "re.compile('[a-z]').findall('abcd')",
            "shutil.get_archive_formats()",
            "tempfile.gettempdir()",
            "uuid.uuid1()",
            "math",
            "[math]",
            "numpy.array([0, 2, 4, 6, 8], dtype='uint8')",
            "np.array([0, 2, 4, 6, 8], dtype='uint8')"
        ]
        for command in module_commands:
            assert pyeval([command])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
from __future__ import (unicode_literals, absolute_import,
                        print_function, division)

from itertools import islice
from signal import signal, SIGPIPE, SIG_DFL

signal(SIGPIPE, SIG_DFL)

import argparse
import collections
import collections.abc
import contextlib
import inspect
import json
import sys, re, io

try:
    from pythonpy.__version__ import __version__
except (ImportError, ValueError, SystemError):
    __version__ = '0.5.2'

pyversion = sys.version.split(' ')[0]
__version_info__ = f'''Pythonpy {__version__}
Python {pyversion}'''

module_aliases = {
    'mp'    : 'matplotlib',
    'np'    : 'numpy',
    'pd'    : 'pandas',
    'tf'    : 'tensorflow',
    'xa'    : 'xarray'
}

ModuleAlias = collections.namedtuple('ModuleAlias', ('shorthand', 'modname'))
IOHandles   = collections.namedtuple('IOHandles', ('out', 'err'))

alias_res   = { re.compile(rf"^{key}") : ModuleAlias(shorthand=key, modname=value) \
                                                                   for key, value \
                                                                    in module_aliases.items() }

def import_matches(query, prefix=''):
    
    for raw_module_name in frozenset(
                           re.findall(
                           rf"({prefix}[a-zA-Z_][a-zA-Z0-9_]*)\.?", query)):
        
        module_name = raw_module_name
        
        for rgx, alias in alias_res.items():
            if rgx.match(module_name):
                module_name = rgx.sub(alias.modname, module_name)
        
        try:
            module = __import__(module_name)
        
        except (ModuleNotFoundError, ImportError):
            pass
        
        else:
            globals()[raw_module_name] = module
            import_matches(query, prefix=f"{module_name}.")

def lazy_imports(*args):
    query = ' '.join([x for x in args if x])
    import_matches(query)

def current_list(input):
    return current_list.rgx.split(input)

current_list.rgx = re.compile(r'[^a-zA-Z0-9_\.]')

def inspect_source(obj):
    import pydoc
    
    try:
        pydoc.pager(''.join(inspect.getsourcelines(obj)[0]))
        return None
    except:
        return help(obj)

parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            add_help=False)

group = parser.add_argument_group("Options")

parser.add_argument('expression', nargs='?', default='None',
                    help="e.g. “py '2 ** 32'”")

group.add_argument('-x', dest='lines_of_stdin', action='store_const',
                    const=True, default=False,
                    help='treat each row of stdin as “x”')

group.add_argument('-fx', dest='filter_result', action='store_const',
                    const=True, default=False,
                    help=argparse.SUPPRESS)

group.add_argument('-l', dest='list_of_stdin', action='store_const',
                    const=True, default=False,
                    help='treat list of stdin as “l”')

group.add_argument('--ji', '--json_input',
                    dest='json_input', action='store_const',
                    const=True, default=False,
                    help=argparse.SUPPRESS)

group.add_argument('--jo', '--json_output',
                    dest='json_output', action='store_const',
                    const=True, default=False,
                    help=argparse.SUPPRESS)

group.add_argument('--si', '--split_input', dest='input_delimiter',
                    help=argparse.SUPPRESS)

group.add_argument('--so', '--split_output', dest='output_delimiter',
                    help=argparse.SUPPRESS)

group.add_argument('-c', dest='pre_cmd', help='run code before expression')
group.add_argument('-C', dest='post_cmd', help='run code after expression')
group.add_argument('--i', '--ignore_exceptions',
                    dest='ignore_exceptions', action='store_const',
                    const=True, default=False,
                    help=argparse.SUPPRESS)

group.add_argument('-V', '--version', action='version', version=__version_info__,
                   help='version info')

group.add_argument('-h', '--help', action='help',
                   help="show this help message and exit")

@contextlib.contextmanager
def redirect(args):
    """ Redirect “stdout” and “stderr” at the same time """
    out, err = io.StringIO(), io.StringIO()
    
    with contextlib.ExitStack() as ctx:
        ctx.enter_context(contextlib.redirect_stdout(out))
        ctx.enter_context(contextlib.redirect_stderr(err))
        iohandles = IOHandles(out=out, err=err)
        
        try:
            yield iohandles
        
        except SystemExit as exit:
            raise TypeError("[ERROR] in cluval execution") from exit
        
        except BaseException as exc:
            import traceback
            pyheader = 'pythonpy/pyeval.py'
            exprheader = 'File "<string>"'
            foundexpr = False
            lines = traceback.format_exception(*sys.exc_info())
            for line in lines:
                if pyheader in line:
                    continue
                iohandles.err.write(line)
                if not foundexpr and line.lstrip().startswith(exprheader) and not isinstance(exc, SyntaxError):
                    iohandles.err.write('    {}\n'.format(args.expression))
                    foundexpr = True
            
            sysexc = SystemExit(iohandles.err.getvalue())
            sysexc.code = 1
            raise sysexc

def print_ok(string):
    try:
        print(string)
    except UnicodeEncodeError:
        print(string.encode('UTF-8'))
        
def safe_eval(code, x):
    try:
        return eval(code)
    except:
        return None

def pyeval(argv=None):
    """ Evaluate a Python expression from a set of CLI arguments. """
    
    args = parser.parse_args(argv)
    
    with redirect(args) as iohandles:
        
        if sum([args.list_of_stdin, args.lines_of_stdin, args.filter_result]) > 1:
            iohandles.err.write('Pythonpy accepts at most one of [-x, -l] flags\n')
            exc = SystemExit(iohandles.err.getvalue())
            exc.code = 1
            raise exc
        
        if args.json_input:
            
            def loads(string):
                try:
                    return json.loads(string.rstrip())
                except BaseException as exc:
                    if args.ignore_exceptions:
                        pass
                    else:
                        raise exc
            
            stdin = (loads(x) for x in sys.stdin)
        
        elif args.input_delimiter:
            stdin = (re.split(args.input_delimiter, x.rstrip()) for x in sys.stdin)
        
        else:
            stdin = (x.rstrip() for x in sys.stdin)
        
        if args.expression:
            args.expression = args.expression.replace("`", "'")
            
            if args.expression.startswith('?') or args.expression.endswith('?'):
                
                final_atom = current_list(args.expression.rstrip('?'))[-1]
                first_atom = current_list(args.expression.lstrip('?'))[0]
                
                if args.expression.startswith('??'):
                    args.expression = f"inspect_source({first_atom})"
                elif args.expression.endswith('??'):
                    args.expression = f"inspect_source({final_atom})"
                elif args.expression.startswith('?'):
                    args.expression = f'help({first_atom})'
                else:
                    args.expression = f'help({final_atom})'
                
                if args.lines_of_stdin:
                    stdin = islice(stdin, 1)
        
        if args.pre_cmd:
            args.pre_cmd = args.pre_cmd.replace("`", "'")
        
        if args.post_cmd:
            args.post_cmd = args.post_cmd.replace("`", "'")
        
        lazy_imports(args.expression, args.pre_cmd, args.post_cmd)
        
        if args.pre_cmd:
            exec(args.pre_cmd)
        
        if args.lines_of_stdin:
            
            if args.ignore_exceptions:
                result = (safe_eval(args.expression, x) for x in stdin)
            else:
                result = (eval(args.expression) for x in stdin)
        
        elif args.filter_result:
            
            if args.ignore_exceptions:
                result = (x for x in stdin if safe_eval(args.expression, x))
            else:
                result = (x for x in stdin if eval(args.expression))
        
        elif args.list_of_stdin:
            
            locals()['l'] = list(stdin)
            result = eval(args.expression)
        
        else:
            result = eval(args.expression)
        
        def prepare(output):
            if output is None:
                return None
            elif args.json_output:
                return json.dumps(output)
            elif args.output_delimiter:
                return args.output_delimiter.join(output)
            else:
                return output
        
        if isinstance(result, collections.abc.Iterable) and not isinstance(result, str):
            for x in result:
                formatted = prepare(x)
                if formatted is not None:
                    print_ok(formatted)
        else:
            formatted = prepare(result)
            if formatted is not None:
                print_ok(formatted)
        
        if args.post_cmd:
            exec(args.post_cmd)
        
        out = iohandles.out.getvalue()
    
    return out



import unittest
from subprocess import check_output

class TestPythonPy(unittest.TestCase):
    
    def checkOutput(self, string, result, **kwargs):
        return self.assertEqual(check_output(string, **kwargs), result)
    
    def test_list_stdin(self):
        self.checkOutput("""echo $'a,2\nb,1' | py -l 'sorted(l, key=lambda x: x.split(",")[1])'""", b'b,1\na,2\n', shell=True)
    
    def test_version(self):
        from pythonpy.__version__ import __version__
        import sys
        pyversion = sys.version.split(' ')[0]
        self.checkOutput(['py', '--version'], bytes(f'''Pythonpy {__version__}\nPython {pyversion}\n''', encoding='UTF-8'))
    
    def test_empty(self):
        self.checkOutput(['py'], b'')
    
    def test_numbers(self):
        self.checkOutput(['py', '3 * 4.5'], b'13.5\n')
    
    def test_range(self):
        self.checkOutput(['py', 'range(3)'], b'0\n1\n2\n')
    
    def test_split_input(self):
        self.checkOutput(["""echo a,b | py -x 'x[1]' --si ,"""], b'b\n', shell=True)
    
    def test_split_output(self):
        self.checkOutput(["""echo abc | py -x x --si '' --so ','"""], b'a,b,c\n', shell=True)
    
    def test_ignore_errors(self):
        self.checkOutput("""echo a | py -x --i 'None.None'""", b'', shell=True)
        self.checkOutput("""echo a | py -fx --i 'None.None'""", b'', shell=True)
    
    def test_statements(self):
        self.checkOutput("""py -c 'a=5' -C 'print(a)'""", b'5\n', shell=True)
        self.checkOutput("""echo 3 | py -c 'a=5' -x x -C 'print(a)'""", b'3\n5\n', shell=True)
    
    def test_imports(self):
        module_commands = ["math.ceil(2.5)",
                           "calendar.weekday(1955, 11, 5)",
                           "csv.list_dialects()",
                           "datetime.timedelta(hours=-5)",
                           "hashlib.sha224(b\"Nobody inspects the spammish repetition\").hexdigest()",
                           "glob.glob('*')",
                           "itertools.product(['a','b'], [1,2])",
                           "json.dumps([1,2,3,{'4': 5, '6': 7}], separators=(',',':'))",
                           "os.name",
                           "random.randint(0, 1000)",
                           "re.compile('[a-z]').findall('abcd')",
                           "shutil.get_archive_formats()",
                           "tempfile.gettempdir()",
                           "uuid.uuid1()",
                           "math",
                           "[math]",
                           "numpy.array([0, 2, 4, 6, 8], dtype='uint8')",
                           "np.array([0, 2, 4, 6, 8], dtype='uint8')",
                           ]
        for command in module_commands:
            check_output("py %r" % command, shell=True)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
from __future__ import (unicode_literals, absolute_import,
                        print_function, division)

from pythonpy.pyeval import pyeval
import sys

def main():
    out = pyeval(argv=sys.argv[1:])
    print(out, end='')


#!/usr/bin/env python
from setuptools import setup
import sys, os

PROJECT_NAME = 'pythonpy'

# PROJECT DIRECTORY
CWD = os.path.dirname(__file__)
BASE_PATH = os.path.join(
            os.path.abspath(CWD), PROJECT_NAME)

py_entry = 'ppy%s = pythonpy.__main__:main'
pycompleter_entry = 'ppycompleter%s = pythonpy.pycompleter:main'
endings = ('', sys.version[:1], sys.version[:3])
entry_points_scripts = []

# PROJECT VERSION & METADATA
__version__ = "<undefined>"
try:
    exec(compile(
        open(os.path.join(BASE_PATH,
            '__version__.py')).read(),
            '__version__.py', 'exec'))
except:
    __version__ = '0.5.2'

for e in endings:
    entry_points_scripts.append(py_entry % e)
    entry_points_scripts.append(pycompleter_entry % e)

setup(
    name='pythonpy-fork',
    version=__version__,
    description='python -c, with tab completion and shorthand; fish2k fork',
    license='MIT',
    url='https://github.com/fish2000/pythonpy-fork',
    long_description='https://github.com/fish2000/pythonpy-fork',
    packages=['pythonpy', 'pythonpy.completion'],
    package_data={'pythonpy': ['completion/pycompletion.sh']},
    scripts=['pythonpy/find_pycompletion.sh'],
    entry_points = {
        'console_scripts': entry_points_scripts
    },
)
